/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef _IBVERBS_HEADER_H
#define _IBVERBS_HEADER_H


#undef IN
#undef OUT

#include <infiniband/verbs.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>

#include "ibverbs_const.h"
#include "via64.h"

double get_us(void);


#define INVAL_HNDL (0xffffffff)

#define IN
#define OUT

#undef MALLOC
#undef FREE
/* src/env/initutil.c NEW not defined */
#define MALLOC(a)    malloc((unsigned)(a))
#define CALLOC(a,b)  calloc((unsigned)(a),(unsigned)(b))
#define FREE(a)      free((char *)(a))
#define NEW(a)    (a *)MALLOC(sizeof(a))
#define STRDUP(a) 	strdup(a)

#if 0
#define D_PRINT(fmt, args...)   {fprintf(stderr, "[%d][%s:%d]", viadev.me, __FILE__, __LINE__);\
                     fprintf(stderr, fmt, ## args); fflush(stderr);}
#else
#define D_PRINT(fmt, args...)
#endif

#define T_PRINT(fmt, args...)

#define DEBUG00 200 /* No debug prints - It is default debug level */
#define DEBUG01 201 
#define DEBUG02 202
#define DEBUG03 203 /* Print All */

/* ch_gen2 layer debug prints are disabled by default. In order
 * to use the macro you need add -DVIADEV_DEBUG to CFLAGS
 */
#ifdef VIADEV_DEBUG
#define V_PRINT (_dlevel, fmt, args...) {                       \
    if (_dlevel <= viadev_debug_level) {                        \
        fprintf(stderr, "[%d][%s:%d]", viadev.me, __FILE__,     \
                __LINE__);                                      \
        fprintf(stderr, fmt, ## args);                          \
        fflush(stderr);                                         \
    }                                                           \
}
#else

#define V_PRINT (_dlevel, fmt, args...)

#endif /* VIADEV_DEBUG */

#if __GNUC__ >= 3
# define VIADEV_LIKELY(x)      __builtin_expect (!!(x), 1)
# define VIADEV_UNLIKELY(x)    __builtin_expect (!!(x), 0)
#else
# define VIADEV_LIKELY(x)      (x)
# define VIADEV_UNLIKELY(x)    (x)
#endif

#endif /* _IBVERBS_HEADER_H */
