/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#define _XOPEN_SOURCE 600

#include "ibverbs_header.h"
#include "mv_buf.h"
#include "mv_priv.h"
#include "dreg.h"
#include "math.h"

mv_sbuf_size mv_sbuf_avail[10];
int mv_sbuf_avail_num = 0;

mv_rbuf_region *rbuf_region_head = NULL;
mv_sbuf_region *sbuf_region_head = NULL;

static mv_rbuf_size mv_rbuf_avail[10];
static int mv_rbuf_avail_num = 0;

static inline mv_sbuf_size * get_sbuf_size(int rbuf_size) {
    int found = 0, i; 
    mv_sbuf_size * b;

    for(i = 0; i < mv_sbuf_avail_num; i++) {
        if(mv_sbuf_avail[i].alloc_size == rbuf_size) {
            found = 1;
            b = &(mv_sbuf_avail[i]);
            break;  
        }       
    }

    if(!found) {
        /* TODO: re-order based on size!!! after we add dynamic allocation  */

        b = &(mv_sbuf_avail[mv_sbuf_avail_num]);

        b->alloc_size = rbuf_size;
        b->max_data_size = rbuf_size - MV_HEADER_OFFSET;
        b->head = NULL; 

        mv_sbuf_avail_num++;
    }

    return b;
}



int release_mv_sbuf(mv_sbuf * v) {
    mv_sbuf_size * b = v->region->size_ptr;

    D_PRINT("releasing %p, current head: %p\n", v, b->head);

    assert(v != b->head);
    assert(v->in_progress == 0);

    v->next_ptr = b->head;
    b->head = v;

    return 0;
}

int release_mv_rbuf(mv_rbuf * v) {
    mv_rbuf_size * b = v->region->size_ptr;

    assert(v != b->head);

    v->next_ptr = b->head;
    b->head = v;

    return 0;
}



static inline mv_rbuf_size * get_rbuf_size(int rbuf_size) {
    int found = 0, i; 
    mv_rbuf_size * b;

    for(i = 0; i < mv_rbuf_avail_num; i++) {
        if(mv_rbuf_avail[i].alloc_size == rbuf_size) {
            found = 1;
            b = &(mv_rbuf_avail[i]);
            break;  
        }       
    }

    if(!found) {
        /* TODO: re-order based on size!!! */

        b = &(mv_rbuf_avail[mv_rbuf_avail_num]);

        b->alloc_size = rbuf_size;
        /* TODO: change if using headers */
        b->max_data_size = rbuf_size;
        b->head = NULL; 

        mv_rbuf_avail_num++;

        D_PRINT("creating new rbuf size: %d\n",  rbuf_size);
    }

    return b;
}


void allocate_mv_rbuf_region(int nvbufs, int rbuf_size)
{
    struct mv_rbuf_region *reg;
    mv_rbuf_size *b;
    void *mem;
    void *vbuf_dma_buffer;

    int i;
    mv_rbuf *cur;
    int alignment_vbuf = 64;
    int alignment_dma;

    alignment_dma = getpagesize();

    assert(nvbufs > 0);

    b = get_rbuf_size(rbuf_size);

    I_PRINT("Alloc rbuf %d size: %d\n", nvbufs, rbuf_size);

    reg = (struct mv_rbuf_region *) malloc(sizeof(struct mv_rbuf_region));
    if (NULL == reg) {
        error_abort_all(GEN_EXIT_ERR,
                "Unable to malloc a new struct mv_rbuf_region");
    }

    if(posix_memalign((void **) &mem, alignment_vbuf, nvbufs * sizeof(mv_rbuf))) {
        error_abort_all(GEN_EXIT_ERR, "unable to malloc vbuf struct");
    }

    if(posix_memalign((void **) &vbuf_dma_buffer, alignment_dma, nvbufs * rbuf_size)) {
        error_abort_all(GEN_EXIT_ERR, "unable to malloc vbufs DMA buffer");
    }

    memset(mem, 0, nvbufs * sizeof(mv_rbuf));
    memset(vbuf_dma_buffer, 0, nvbufs * rbuf_size);

    reg->malloc_start = mem;
    reg->malloc_end = (void *) ((char *) mem + nvbufs * sizeof(mv_rbuf));
    reg->malloc_buf_start = vbuf_dma_buffer;
    reg->malloc_buf_end = (void *) ((char *) vbuf_dma_buffer + nvbufs * rbuf_size);
    reg->buf_count = nvbufs;
    reg->size_ptr = b;

    b->head = reg->head = (mv_rbuf *) ((aint_t) mem);

    for(i = 0; i < mvdev.num_hcas; i++) {
        reg->mem_handle[i] = register_memory(i, vbuf_dma_buffer, nvbufs * rbuf_size);
    }

    /* init the free list */
    for (i = 0; i < nvbufs - 1; i++) {
        cur = reg->head + i;
        cur->next_ptr = reg->head + i + 1;
        cur->region = reg;
        cur->header_ptr = cur->base_ptr = (unsigned char *) ((unsigned char *)(vbuf_dma_buffer) +
                (i * rbuf_size));
        cur->desc.parent = cur;
    }


    /* last one needs to be set to NULL */
    cur = reg->head + nvbufs - 1;
    cur->next_ptr = NULL;
    cur->region = reg;
    cur->header_ptr = cur->base_ptr = (unsigned char *) ((unsigned char *)vbuf_dma_buffer + 
            ((nvbufs - 1) * rbuf_size));
    cur->desc.parent = cur;


    /* thread region list */
    reg->next = rbuf_region_head;
    rbuf_region_head = reg;

}

#define INIT_DESC(_v, _num, _i) { \
    int _z = 0; \
    for(_z = 0; _z < _num; _z++) { \
        if(_v->desc[_z].parent != NULL) { \
            fprintf(stderr, "z: %d, i: %d\n", _z, _i); \
            assert(_v->desc[_z].parent == NULL); \
        } \
        _v->desc[_z].parent = _v; \
    } \
}

void allocate_mv_sbuf_region(int nvbufs, int sbuf_size)
{
    struct mv_sbuf_region *reg;
    mv_sbuf_size *b;
    void *mem;
    void *vbuf_dma_buffer;
    void *desc_mem_buffer;
    mv_sdescriptor *desc_mem;

    int i;
    mv_sbuf *cur = NULL;
    int alignment_vbuf = 64;
    int alignment_dma;
    int desc_num = ceil((double) sbuf_size / (mvparams.mtu - 40));

    alignment_dma = getpagesize();

    I_PRINT("Alloc sbuf %d size: %d\n", nvbufs, sbuf_size);

    assert(nvbufs > 0);

    b = get_sbuf_size(sbuf_size);

    reg = (struct mv_sbuf_region *) malloc(sizeof(struct mv_sbuf_region));
    if (NULL == reg) {
        error_abort_all(GEN_EXIT_ERR,
                "Unable to malloc a new struct mv_sbuf_region");
    }

    if(posix_memalign((void **) &mem, alignment_vbuf, nvbufs * sizeof(mv_sbuf))) {
        error_abort_all(GEN_EXIT_ERR, "unable to malloc vbuf struct");
    }

    if(posix_memalign((void **) &vbuf_dma_buffer, alignment_dma, nvbufs * sbuf_size)) {
        error_abort_all(GEN_EXIT_ERR, "unable to malloc vbufs DMA buffer");
    }

    if(posix_memalign((void **) &desc_mem_buffer, alignment_vbuf, 
                nvbufs * sizeof(mv_sdescriptor) * desc_num)) {
        error_abort_all(GEN_EXIT_ERR, "unable to malloc vbufs DMA buffer");
    }

    desc_mem = (mv_sdescriptor *) desc_mem_buffer;

    memset(mem, 0, nvbufs * sizeof(mv_sbuf));
    memset(vbuf_dma_buffer, 0, nvbufs * sbuf_size);
    memset(desc_mem, 0, nvbufs * sizeof(mv_sdescriptor) * desc_num);

    reg->malloc_start = mem;
    reg->malloc_end = (void *) ((char *) mem + nvbufs * sizeof(mv_sbuf));
    reg->malloc_buf_start = vbuf_dma_buffer;
    reg->malloc_buf_end = (void *) ((char *) vbuf_dma_buffer + nvbufs * sbuf_size);

    reg->malloc_desc_start = desc_mem;
    reg->malloc_desc_end = (void *) ((char *) desc_mem + nvbufs * sizeof(mv_sdescriptor) * desc_num);

    reg->buf_count = nvbufs;
    reg->size_ptr = b;

    b->head = reg->head = (mv_sbuf *) ((aint_t) mem);

    D_PRINT("head is %p\n", b->head);

    for(i = 0; i < mvdev.num_hcas; i++) {
        reg->mem_handle[i] = register_memory(i, vbuf_dma_buffer, nvbufs * sbuf_size);
    }
    D_PRINT("reg lkey: %p\n", reg->mem_handle[0]->lkey);

    /* init the free list */
    for (i = 0; i < nvbufs - 1; i++) {
        cur = reg->head + i;
        cur->next_ptr = reg->head + i + 1;
        cur->region = reg;
        cur->base_ptr = (unsigned char *) ((unsigned char *)(vbuf_dma_buffer) +
                (i * sbuf_size));
        cur->desc = &(desc_mem[i * desc_num]);
        cur->in_progress = 0;
        
        INIT_DESC(cur, desc_num, i);
    }

    cur->next_ptr = NULL;

    /* thread region list */
    reg->next = sbuf_region_head;
    sbuf_region_head = reg;

}

mv_rbuf *get_mv_rbuf(int length_requested) {
    int i;  
    mv_rbuf *v = NULL; 

    for(i = 0; i < mv_rbuf_avail_num; i++) {
        if(mv_rbuf_avail[i].max_data_size >= length_requested ||
                i == (mv_rbuf_avail_num - 1)) {
            /* get the head from this list... */

            if(NULL == mv_rbuf_avail[i].head) {
                allocate_mv_rbuf_region(256, mv_rbuf_avail[i].alloc_size);
            }

            v = mv_rbuf_avail[i].head;
            v->header_ptr = v->base_ptr + 40 + MV_HEADER_OFFSET; 
            v->mheader_ptr = v->base_ptr + 40;

            mv_rbuf_avail[i].head = (struct mv_rbuf *) v->next_ptr;

            return v;
        }
    }
    assert(0);
    return NULL;
}

mv_sbuf *get_mv_sbuf(int length_requested) {
    int i;  
    mv_sbuf *v = NULL; 

    D_PRINT("len requested: %d, avail: %d\n", length_requested, mv_sbuf_avail_num);

    for(i = 0; i < mv_sbuf_avail_num; i++) {
        if(mv_sbuf_avail[i].max_data_size >= length_requested ||
                i == (mv_sbuf_avail_num - 1)) { 
            /* get the head from this list... */

            if(NULL == mv_sbuf_avail[i].head) {
                D_PRINT("head is null, so allocate\n");
                allocate_mv_sbuf_region(128, mv_sbuf_avail[i].alloc_size);
            }       

            v = mv_sbuf_avail[i].head;

            v->header_ptr = v->base_ptr + MV_HEADER_OFFSET;
            v->left_to_send = 0;

            mv_sbuf_avail[i].head = (struct mv_sbuf *) v->next_ptr;

            v->in_sendwin = 0;
            v->segments = 0;
            v->retry_count = 0;
            v->in_progress = 1;
            v->retry_always = 0;

            return v;
        }       
    }

    assert(0);
    return NULL;
}

void allocate_mv_bufs(int nvbufs)
{
    allocate_mv_rbuf_region(1024, mvparams.mtu);

    allocate_mv_sbuf_region(128, 512);
    if(mvparams.mtu > 2048) {
        allocate_mv_sbuf_region(128, 2048);
    }
    allocate_mv_sbuf_region(128, mvparams.mtu);

    if(!mvparams.use_headers) {
        allocate_mv_sbuf_region(64, 2048 * 2);
        allocate_mv_sbuf_region(64, 2048 * 9);
        allocate_mv_sbuf_region(32, 2048 * 8);
    }
}


void deallocate_mv_bufs() {

    int i = 0;
    mv_rbuf_region *r_reg = rbuf_region_head;
    mv_sbuf_region *s_reg = sbuf_region_head;

    while (r_reg) {
        for(i = 0; i < mvdev.num_hcas; i++) {
            if(r_reg->mem_handle[i] != NULL) {
                if(ibv_dereg_mr(r_reg->mem_handle[i])) {
                    fprintf(stderr, "[%d] Error in deregistration\n", mvdev.me); 
                }
            }
        }

        free(r_reg->malloc_start);
        free(r_reg->malloc_buf_start);

        r_reg = r_reg->next;
    }

    while (s_reg) {
        for(i = 0; i < mvdev.num_hcas; i++) {
            if(s_reg->mem_handle != NULL) {
                if(ibv_dereg_mr(s_reg->mem_handle[i])) {
                    fprintf(stderr, "[%d] Error in deregistration\n", mvdev.me); 
                }
            }
        }

        free(s_reg->malloc_start);
        free(s_reg->malloc_buf_start);
        free(s_reg->malloc_desc_start);

        s_reg = s_reg->next;
    }

}


