/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#ifndef _VBUF_H
#define _VBUF_H

#include <pthread.h>
#include "ibverbs_header.h"
#include "mv_param.h"
#include "get_clock.h"

typedef enum {
   MVDEV_RPUT_FLAG,
   MVDEV_RGET_FLAG,
} mv_flags;

typedef struct mv_sdescriptor {
    struct ibv_send_wr sr;
    struct ibv_sge sg_entry;
    void *parent;
    struct mv_sdescriptor * next_extsendq;
    void * qp;
} mv_sdescriptor;

typedef struct mv_rdescriptor {
    struct ibv_recv_wr rr;
    struct ibv_sge sg_entry;
    void *parent;
    void * qp;
} mv_rdescriptor;

typedef struct listptr {
    void *next;
    void *prev;
} listptr;

typedef struct mv_sbuf {
    unsigned char *base_ptr;
    unsigned char *header_ptr;
    mv_sdescriptor * desc;
    struct mv_sbuf_region *region;

    listptr extwin_ptr;
    listptr sendwin_ptr;
    listptr unackq_ptr;
    void * next_ptr;

    uint32_t seqnum;
    uint32_t seqnum_last;

    uint32_t rank;

    uint8_t left_to_send;
    uint8_t retry_always;
    uint8_t in_sendwin;
    uint8_t transport;
    uint8_t in_progress;
    uint16_t segments;
    uint16_t retry_count;

    void *shandle;
    double timestamp;
} mv_sbuf;

typedef struct mv_rbuf {
    unsigned char *base_ptr;
    unsigned char *mheader_ptr;
    unsigned char *header_ptr;
    unsigned char *data_ptr;
    mv_rdescriptor desc;
    struct mv_rbuf_region *region;

    uint32_t byte_len;
    uint32_t byte_len_data;

    listptr recvwin_ptr;
    listptr eager_ptr;
    void * next_ptr;

    uint32_t seqnum;
    uint32_t rank;
    uint8_t has_header;
    uint8_t transport;
} mv_rbuf;


typedef struct mv_sbuf_size {
    uint32_t max_data_size;
    uint32_t alloc_size;
    struct mv_sbuf * head;
    listptr next;
} mv_sbuf_size;

typedef struct mv_rbuf_size {
    uint32_t max_data_size;
    uint32_t alloc_size;
    struct mv_rbuf * head;
    listptr next;
} mv_rbuf_size;



typedef struct mv_rbuf_region {
    struct ibv_mr *mem_handle[MAX_NUM_HCAS];  /* memory handle for entire region */
    int buf_size;               /* how many bytes each vbuf in this region is */

    void *malloc_start;         /* used to free region later */
    void *malloc_end;           /* to bracket mem region */
    void *malloc_buf_start;     /* used to free DMA region later */
    void *malloc_buf_end;       /* bracket DMA region */
    int buf_count;              /* number of vbufs in region */
    struct mv_rbuf *head;       /* first vbuf in region */
    struct mv_rbuf_region *next;   /* thread vbuf regions */
    struct mv_rbuf_size *size_ptr;
} mv_rbuf_region;

typedef struct mv_sbuf_region {
    struct ibv_mr *mem_handle[MAX_NUM_HCAS];  /* memory handle for entire region */
    int buf_size;               /* how many bytes each vbuf in this region is */

    void *malloc_start;         /* used to free region later */
    void *malloc_end;           /* to bracket mem region */
    void *malloc_buf_start;     /* used to free DMA region later */
    void *malloc_buf_end;       /* bracket DMA region */
    int buf_count;              /* number of vbufs in region */
    struct mv_sbuf *head;       /* first vbuf in region */
    struct mv_sbuf_region *next;   /* thread vbuf regions */
    struct mv_sbuf_size *size_ptr;

    void *malloc_desc_start;
    void *malloc_desc_end;
} mv_sbuf_region;

#define MV_HEADER_OFFSET (mvparams.use_headers ? sizeof(mvdev_packet_mheader) : 0)
#define MV_TRANSPORT_OFFSET(_v) ((_v->transport == MVDEV_TRANSPORT_UD) ? 40 : 0)
#define SBUF_MAX_DATA_SZ(_v) (_v->region->size_ptr->max_data_size - MV_TRANSPORT_OFFSET(_v))
#define SBUF_MAX_DATA_SZ_HDR(_v, type) (_v->region->size_ptr->max_data_size - sizeof(type) - MV_TRANSPORT_OFFSET(_v))

#define MV_SBUF_SET_DATA(_v, type) { \
    _v->data_ptr = _v->header_ptr + sizeof(type);  \
    _v->byte_len_data = _v->byte_len - sizeof(type); \
}


void MV_Allocate_Vbufs(int nvbufs);


void deallocate_mv_bufs(void);
void allocate_mv_bufs(int nvbufs);
mv_sbuf *get_mv_sbuf(int length_requested);
mv_rbuf *get_mv_rbuf(int length_requested);
int release_mv_sbuf(mv_sbuf * v);
int release_mv_rbuf(mv_rbuf * v);

void init_vbuf_lock();
void check_rbufs();

void allocate_mv_rbuf_region(int nvbufs, int rbuf_size);
void allocate_mv_sbuf_region(int nvbufs, int sbuf_size);

#endif                          /* _VBUF_H */
