/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#ifndef _MV_INLINE_H
#define _MV_INLINE_H

#include <stdio.h>
#include <unistd.h>
#include "mpid.h"
#include "ibverbs_header.h"
#include "mv_dev.h"
#include "mv_priv.h"
#include "mv_buf.h"
#include "mv_packet.h"

static inline void mvdev_post_channel_send_qp_lid(mvdev_connection_t * c, mv_sbuf * v, int total_len, mv_qp *qp, int lid) {
        mvdev_func.post_ud_send(c, v, total_len, qp, 1, lid); 
}

static inline void mvdev_post_channel_send_qp(mvdev_connection_t * c, mv_sbuf * v, int total_len, mv_qp *qp) {
        mvdev_func.post_ud_send(c, v, total_len, qp, 1, -1); 
}

static inline void mvdev_post_channel_send_qp_ns(mvdev_connection_t * c, mv_sbuf * v, int total_len, mv_qp *qp) {
        mvdev_func.post_ud_send(c, v, total_len, qp, 0, -1); 
}

static inline void mvdev_post_channel_send(mvdev_connection_t * c, mv_sbuf * v, int total_len) {
        mvdev_func.post_ud_send(c, v, total_len, NULL, 1, -1);
}



static inline void mvdev_ext_sendq_queue(mv_qp *qp, mv_sdescriptor * d)
{
    qp->ext_sendq_size++;
    d->next_extsendq = NULL;
    if(qp->ext_sendq_head == NULL) {
        qp->ext_sendq_head = d;
    } else {    
        qp->ext_sendq_tail->next_extsendq = d;
    }

    qp->ext_sendq_tail = d;
    D_PRINT("extsq\n");
}

#define SET_MHEADER(_mp, _seqnum, _hasheader) {        \
    (_mp)->has_header = _hasheader;                    \
    (_mp)->seqnum = _seqnum;                           \
    (_mp)->src_rank = mvdev.me;                        \
}

static inline void prepare_ud_descriptor_noheaders(mv_sdescriptor *desc, unsigned char * addr,
        mv_sbuf *sbuf,
        unsigned len, int peer, uint32_t seqnum, uint32_t has_header,
        mv_qp * qp, int last, int ah_index)
{
    desc->sr.next = NULL;
    if(len <= mvparams.max_ud_inline) {
        desc->sr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE; 
    } else {
        desc->sr.send_flags = IBV_SEND_SIGNALED;
    }
    desc->sr.opcode = IBV_WR_SEND;
    desc->sr.wr_id = (aint_t) desc;
    desc->sr.num_sge = 1;
    desc->sr.sg_list = &(desc->sg_entry);

    if(ah_index != -1) {
        desc->sr.wr.ud.ah = mvdev.connections[peer].data_ud_ah[ah_index];
    } else {
        desc->sr.wr.ud.ah = mvdev.connections[peer].data_ud_ah[mvdev.connections[peer].last_ah];
        mvdev.connections[peer].last_ah = 
            (mvdev.connections[peer].last_ah + 1) % mvparams.max_ah_total;
    }
    desc->sr.wr.ud.remote_qpn = mvdev.qpns[peer];
    desc->sr.wr.ud.remote_qkey = 0;

    desc->sg_entry.addr = (uintptr_t) addr;
    desc->sg_entry.length = len;
    desc->sg_entry.lkey = sbuf->region->mem_handle[0]->lkey;

    desc->qp = qp;
}
static inline void prepare_ud_descriptor(mv_sdescriptor *desc, unsigned char * addr,
        mv_sbuf *sbuf,
        unsigned len, int peer, uint32_t seqnum, uint32_t has_header,
        mv_qp * qp, int last, int ah_index)
{
    desc->sr.next = NULL;

    if(MVDEV_LIKELY(last)) {
        if(len <= mvparams.max_ud_inline) {
            desc->sr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE; 
        } else {
            desc->sr.send_flags = IBV_SEND_SIGNALED;
        }
    } else {
        if(len <= mvparams.max_ud_inline) {
            desc->sr.send_flags = IBV_SEND_INLINE; 
        } else {
            desc->sr.send_flags = 0;
        }
    }

    desc->sr.opcode = IBV_WR_SEND_WITH_IMM;
    desc->sr.wr_id = (aint_t) desc;
    desc->sr.num_sge = 1;
    desc->sr.sg_list = &(desc->sg_entry);

    desc->sr.imm_data = CREATE_THS_TUPLE(has_header, mvdev.me, seqnum);

    if(ah_index != -1) {
        desc->sr.wr.ud.ah = mvdev.connections[peer].data_ud_ah[ah_index];
    } else {
        desc->sr.wr.ud.ah = mvdev.connections[peer].data_ud_ah[mvdev.connections[peer].last_ah];
        mvdev.connections[peer].last_ah = 
            (mvdev.connections[peer].last_ah + 1) % mvparams.max_ah_total;
    }

    desc->sr.wr.ud.remote_qpn = mvdev.qpns[peer];
    desc->sr.wr.ud.remote_qkey = 0;

    desc->sg_entry.addr = (uintptr_t) addr;
    desc->sg_entry.length = len;
    desc->sg_entry.lkey = sbuf->region->mem_handle[0]->lkey;

    desc->qp = qp;
}



static inline void mvdev_post_channel_send_control(mvdev_connection_t * c, mv_sbuf * v, unsigned total_len) {
    struct ibv_send_wr *bad_wr;

    mvdev.last_qp = (mvdev.last_qp + 1) % mvdev.num_ud_qps;

    v->seqnum = 0;
    v->segments = 1;
    v->left_to_send = 1;

    if(mvparams.use_headers) {
        mvdev_packet_mheader * mp = (mvdev_packet_mheader *) v->base_ptr;
        SET_MHEADER(mp, 0, 1);
        prepare_ud_descriptor_noheaders( &(v->desc[0]), v->base_ptr,
            v, total_len + sizeof(mvdev_packet_mheader), c->global_rank, 
            v->seqnum, 1, &(mvdev.ud_qp[mvdev.last_qp]), 1, -1);

    } else {
        prepare_ud_descriptor( &(v->desc[0]), v->base_ptr,
            v, total_len, c->global_rank, v->seqnum, 1, &(mvdev.ud_qp[mvdev.last_qp]), 1, -1);
    }

    mvdev.connection_ack_required[c->global_rank] = 0;

    if(mvdev.ud_qp[mvdev.last_qp].send_wqes_avail <= 0) {
        mvdev_ext_sendq_queue(&(mvdev.ud_qp[mvdev.last_qp]), &(v->desc[0]));
    } else {
        mvdev.ud_qp[mvdev.last_qp].send_wqes_avail--;

        if(ibv_post_send(mvdev.ud_qp[mvdev.last_qp].qp, &(v->desc[0].sr), &bad_wr)) {
            error_abort_all(IBV_RETURN_ERR,"Error posting to UD QP %d\n",
                  mvdev.ud_qp[mvdev.last_qp].send_wqes_avail);
        }
    }
}

/* we only add to the end of the sendwin, making this a bit easier... */

static inline void mvdev_sendwin_add(mvdev_connection_t * c, mv_sbuf *ptr)  {
    ptr->sendwin_ptr.next = ptr->sendwin_ptr.prev = NULL;

    ptr->in_sendwin = 1;

    if(NULL == c->send_window_head) {
        c->send_window_head = ptr;
    } else {
        (c->send_window_tail)->sendwin_ptr.next = ptr;
    }

    c->send_window_tail = ptr;
    c->send_window_size++;
}

static inline void mvdev_unackq_add(mv_sbuf *ptr)  {
    ptr->unackq_ptr.next = NULL; 

    if(NULL == mvdev.unack_queue_head) {
        mvdev.unack_queue_head = ptr;
        ptr->unackq_ptr.prev = NULL; 
    } else {
        mvdev.unack_queue_tail->unackq_ptr.next = ptr;
        ptr->unackq_ptr.prev = mvdev.unack_queue_tail;
    }

    mvdev.unack_queue_tail = ptr;
}

static inline void mvdev_track_send(mvdev_connection_t * c, mv_sbuf * v) {

    assert(v->seqnum != 0);


    if(!v->in_sendwin) {
        D_PRINT("adding to swin: %p\n", v);
        v->timestamp = get_time_us();

        mvdev_sendwin_add(c, v);
        mvdev.connection_ack_required[c->global_rank] = 0;
        mvdev_unackq_add(v);

    }

    c->send_window_segments++;
}

/* we can remove from anywhere within the unackq */
static inline void mvdev_unackq_remove(mv_sbuf *ptr)  {
    mv_sbuf * prev_ptr = (mv_sbuf *) ptr->unackq_ptr.prev;
    mv_sbuf * next_ptr = (mv_sbuf *) ptr->unackq_ptr.next;

    if(NULL == prev_ptr) {
        mvdev.unack_queue_head = next_ptr;
    } else {
        prev_ptr->unackq_ptr.next = next_ptr;
    }

    if(NULL == next_ptr) {
        mvdev.unack_queue_tail = prev_ptr;
    } else {
        next_ptr->unackq_ptr.prev = prev_ptr;
    }
}

static inline void mvdev_unackq_traverse() {
    double timestamp;
    int delay;
    mv_sbuf * current;

    current = mvdev.unack_queue_head;

    timestamp = get_time_us();

    while(NULL != current) {

        if(0 == current->left_to_send || current->retry_always) {
            delay = timestamp - current->timestamp;

            if(delay > mvparams.retry_usec) {
                current->timestamp = timestamp;
                mvdev_resend(current);
                timestamp = get_time_us();
            }
        }

        current = current->unackq_ptr.next;
    }
}

static inline void mvdev_ext_sendq_send(mv_qp * qp)
{
    mv_sdescriptor *d;
    struct ibv_send_wr *bad_wr;
    struct ibv_send_wr *sr;
    int i = 0;

    while (qp->send_wqes_avail > 0 && qp->ext_sendq_head) {
        d = qp->ext_sendq_head;

        /* find how many desc are chained */
        i = 1;
        sr = &(d->sr);
        while(sr->next) {
            sr = sr->next;
            i++;
        }

        if(qp->send_wqes_avail >= i) {
            qp->ext_sendq_size--;

            qp->ext_sendq_head = d->next_extsendq;
            if (d == qp->ext_sendq_tail) {
                qp->ext_sendq_tail = NULL; 
            }       
            d->next_extsendq = NULL; 
            qp->send_wqes_avail -= i;

            if(ibv_post_send(qp->qp, &(d->sr), &bad_wr)) {
                error_abort_all(IBV_RETURN_ERR,"Error posting to UD QP\n");
            }
        } else {
            break;
        }
    }
}

static inline void mvdev_recvwin_add(mvdev_connection_t * c, mv_rbuf *ptr)  {
    ptr->recvwin_ptr.next = ptr->recvwin_ptr.prev = NULL;

    if(MVDEV_UNLIKELY(NULL == c->recv_window_head)) {
        c->recv_window_head = ptr;
        c->recv_window_tail = ptr;
    } else {
        mv_rbuf * curr_rbuf = c->recv_window_head;

        if(MVDEV_LIKELY(ptr->seqnum > c->recv_window_start_seq)) {
            if(curr_rbuf->seqnum < c->recv_window_start_seq) {
            } 
            else {
                while(NULL != curr_rbuf && curr_rbuf->seqnum < ptr->seqnum
                        && curr_rbuf->seqnum > c->recv_window_start_seq) {
                    curr_rbuf = curr_rbuf->recvwin_ptr.next;
                }
            }
        } else {
            if(MVDEV_UNLIKELY(curr_rbuf->seqnum > c->recv_window_start_seq)) {

                while(NULL != curr_rbuf &&
                        ((curr_rbuf->seqnum >= c->recv_window_start_seq) ||
                         (curr_rbuf->seqnum < ptr->seqnum))) {
                    curr_rbuf = curr_rbuf->recvwin_ptr.next;
                }
            } else {
                while(NULL != curr_rbuf && curr_rbuf->seqnum < ptr->seqnum) {
                    curr_rbuf = curr_rbuf->recvwin_ptr.next;
                }
            }
        }

        if(MVDEV_LIKELY(NULL != curr_rbuf)) {
            if(curr_rbuf->seqnum == ptr->seqnum) {
                return; /* a bit hacky to just return, but... */
            }

            ptr->recvwin_ptr.next = curr_rbuf;
            ptr->recvwin_ptr.prev = curr_rbuf->recvwin_ptr.prev;

            if(curr_rbuf == c->recv_window_head) {
                c->recv_window_head = ptr;
            } else {
                ((mv_rbuf *) curr_rbuf->recvwin_ptr.prev)->recvwin_ptr.next = ptr;
            }
            curr_rbuf->recvwin_ptr.prev = ptr;
        } else {
            ptr->recvwin_ptr.next = NULL;
            ptr->recvwin_ptr.prev = c->recv_window_tail;
            c->recv_window_tail->recvwin_ptr.next = ptr;
            c->recv_window_tail = ptr;
        }
    }
    c->recv_window_size++;
}


static inline void mvdev_windowq_send(mvdev_connection_t * c)
{
    mv_sbuf * next;
    while(NULL != c->ext_window_head && c->send_window_segments <= mvparams.send_window_segments_max) {
        next = c->ext_window_head->extwin_ptr.next;
        mvdev_post_channel_send(c, c->ext_window_head, c->ext_window_head->seqnum);
        c->ext_window_head->extwin_ptr.next = NULL;
        c->ext_window_head = next;
    }

    if(c->ext_window_head == NULL) {
        c->ext_window_tail = NULL;
    }
}

static inline void mvdev_sendwin_remove(mvdev_connection_t * c, mv_sbuf *ptr)  {
    mv_sbuf * next_ptr = (mv_sbuf *) ptr->sendwin_ptr.next;

    ptr->in_sendwin = 0;

    c->send_window_head = next_ptr;
    c->send_window_size--;
    c->send_window_segments -= ptr->segments;

    /* see if we can flush anything from the window queue */
    if(NULL != c->ext_window_head &&
            c->send_window_segments < mvparams.send_window_segments_max) {
        mvdev_windowq_send(c);
    } 

    if(c->blocked_count > 0) {
        PUSH_FLOWLIST(c);
    }
}

static inline void mvdev_recvwin_remove(mvdev_connection_t * c, mv_rbuf *ptr)  {
    mv_rbuf * next_ptr = (mv_rbuf *) ptr->recvwin_ptr.next;

    c->recv_window_head = next_ptr;
    if(next_ptr != NULL) {
        next_ptr->recvwin_ptr.prev = NULL;
    } else {
        c->recv_window_tail = NULL;
        c->recv_window_head = NULL;
    }

    c->recv_window_size--;
}


/* Figure out where in the list we need to add this packet. 
 * If we are the first entry then there is no need to queue the packet
 * and it can just be processed. That will be the quickest path
 */

static inline void mvdev_place_recvwin(mvdev_connection_t * c, mv_rbuf * v) {
    packet_sequence_t win_start = c->recv_window_start_seq;
    packet_sequence_t win_end   =
        (c->recv_window_start_seq + mvparams.recv_window_size_max) % MAX_SEQ_NUM;

    /* check if the packet is in the window or not */
    D_PRINT("Entering mvdev_place_recvwin\n");

    if(MVDEV_LIKELY(INCL_BETWEEN(v->seqnum, win_start, win_end))) {

        if(MVDEV_LIKELY(v->seqnum == c->recv_window_start_seq)) {

            D_PRINT("Got exact in-order recv (got %d)\n", v->seqnum);

            /* we are next in the queue to be processes */
            mvdev_process_recv_inorder(c, v);
            mvdev.connection_ack_required[c->global_rank] = 1;
            c->seqnum_next_toack = v->seqnum;
            INCREMENT_SEQ_NUM(&(c->recv_window_start_seq));
        }
        else {
            /* we are not in order... */
            D_PRINT("Got out-of-order recv (got %d, exp: %d)\n", v->seqnum, win_start);
            mvdev_recvwin_add(c, v);
        }

        while(NULL != c->recv_window_head &&
                PEEK_RECV_WIN(c) == c->recv_window_start_seq) {
            int current_seqnum = PEEK_RECV_WIN(c);
            mvdev_process_recv_inorder(c, c->recv_window_head);
            mvdev_recvwin_remove(c, c->recv_window_head);
            c->seqnum_next_toack = current_seqnum;

            INCREMENT_SEQ_NUM(&(c->recv_window_start_seq));
        }

    } else {
        D_PRINT("Message is not in window, seq: %d, start: %d, end: %d\n", 
                v->seqnum, win_start, win_end);
        release_mv_rbuf(v);
    }

    mvdev.connection_ack_required[c->global_rank] = 1;
}

static inline void mvdev_process_ack(mvdev_connection_t *c, uint32_t last_recv)
{
    D_PRINT("Last recv'd pkt (from remote): %d, last acked: %d\n",
            last_recv, c->seqnum_next_toack);

    mv_sbuf * head_ptr = c->send_window_head;
    mv_sbuf * tail_ptr = c->send_window_tail;

    while(NULL != head_ptr &&
            INCL_BETWEEN(last_recv, head_ptr->seqnum, tail_ptr->seqnum_last)) {

        if(MVDEV_LIKELY(INCL_BETWEEN(head_ptr->seqnum_last, head_ptr->seqnum, last_recv))) {
            D_PRINT("freeing from sendwin!\n");

            mvdev_sendwin_remove(c, head_ptr);
            mvdev_unackq_remove(head_ptr);
            mvdev_process_send_afterack(head_ptr);

            head_ptr = c->send_window_head;

        } 
        else {
            /* only got a partial message acknowledgement, we
             * can try to do more intelligent things here. e.g.
             * it is somewhat likely the end of the message got dropped
             */
            D_PRINT("Got partial! start: %d, end: %d, last: %d\n", head_ptr->seqnum, head_ptr->seqnum_last, last_recv);
            break;
        }
    }
}



static inline void mvdev_explicit_ack(int rank) {
    mvdev_connection_t *c = &(mvdev.connections[rank]);
    mv_sbuf *v = get_sbuf(c, sizeof(mvdev_packet_basic));
    mvdev_packet_basic * p = (mvdev_packet_basic *) v->header_ptr;

    assert(rank != mvdev.me);

    PACKET_SET_HEADER(p, c, MVDEV_PACKET_ACK);
    mvdev_post_channel_send_control(c, v, sizeof(mvdev_packet_basic));

    assert(p->header.type == MVDEV_PACKET_ACK);

    D_PRINT("Sending explict ACK to %d\n", rank);
}


static inline void mvdev_send_acks() {
    int i;
 
    /* there are a few ways of doing this -- probably creating
     * a new array with ack required is best on a per-connection
     * basis. This could get messy with thousands of processors
     * though -- so maybe a linked list is best
     */
 
    for(i = 0; i < mvdev.np; i++) {
        if(mvdev.connection_ack_required[i]) {
            mvdev_explicit_ack(i);
        }
    }
    
}

#define SEND_ACKS() { \
    int ack_i; \
    for(ack_i = 0; ack_i < mvdev.np; ack_i++) { \
        if(mvdev.connection_ack_required[ack_i]) { \
            mvdev_explicit_ack(ack_i); \
        }       \
    } \
}

static inline void mvdev_windowq_queue(mvdev_connection_t * c, mv_sbuf * v, int total_len)
{
    static int i = 0;
    v->seqnum = total_len;
    v->extwin_ptr.next = NULL;
    if (c->ext_window_head == NULL) {
        c->ext_window_head = v;
    } else {    
        c->ext_window_tail->extwin_ptr.next = v;
    }
    c->ext_window_tail = v;

    i++;
    if(i == 5) {
        mvdev_explicit_ack(c->global_rank);
        i = 0;
    }
}

#endif
